#ifndef __FINITE_STATE_MACHINE_H__
#define __FINITE_STATE_MACHINE_H__

// Loki primitives, a few types ported from Loki library. 
// Here you can use Loki original sources instead of #include "LokiPort.h" statement
#include "LokiPort.h"


namespace fsm    // finite state mashine namespace
{

////////////////////////////////////////////////////////////////////////////////
// class template MultipleInheritance
// Let you generate own class derived (multiple inheritance) from each type belongs to typelist
// Invocation (TList is a typelist):
// class MyClass : public MultipleInheritance<TYPELIST_3(Type1,Type2,Type3)> {};
// result:
//      Type1       Type2       Type3
//        ^           ^           ^
//        |___________|___________|
//                    |
//                  MyClass
// where MyClass is derived class from Type1, Type2 and Type3.
////////////////////////////////////////////////////////////////////////////////

template<typename TList>
struct MultipleInheritance : public TList::Head, public MultipleInheritance<typename TList::Tail>
{
};

template<>
struct MultipleInheritance<Loki::NullType>
{
};


//////////////////////////////////////////////////////////////
// Transition template generator
// In order to define your own transition type use the following C++ construct
// typedef TransitionT<CONTEXT_T>::DefineT<BaseState,Event,TargetState,CONTEXT_T::Action> myTransition;
// where
//  - CONTEXT_T is an user defined type (class) wich encapsulates all or part activities for particular State machine.
//  - CONTEXT_T::Action - represents member function with predefined signature : void* (*)(); This method will be invoked after TargetState is set.
//  - STATE_FROM        - an integer constant that uniquely identifies the BEGIN state for given transition. 
//  - STATE_TO          - an integer constant that uniquely identifies the END state for given transition. 
//  - EVENT_VALUE       - an integer constant that uniquely identifies the concrete event which triggers given transition.
// Note: 
// To simplify transition definitions you can use FSM_TRANSITION and FSM_TRANSITION_NO_ACTION macros. Please see below.
template
<
    typename CONTEXT_T
>
class TransitionT
{
    typedef void (CONTEXT_T::*ActionT)();   // predefined signature of member function: void* (*)()
    template<ActionT ACTION_ARG> struct MemberFunction2Type {};
    
public:
    template
    <
        int                 STATE_FROM,
        int                 EVENT_VALUE,
        int                 STATE_TO,
        ActionT             ACTION
    >
    struct DefineT
    {
        typedef CONTEXT_T ContextT;

        enum { BASE_STATE = STATE_FROM, TARGET_STATE = STATE_TO, EVENT = EVENT_VALUE };
        static void    invoke(ContextT* context)    { (*context.*ACTION)(); }

    private:
        // CHECK_DUPLICATION: to avoid programmer's mistake when the same transition (pair: 'StateFrom and Event' identifies the particular transition uniquely)
        // 1.   can invoke different actions
        // 2.   can set different state-destinations
        // the following transition table definition
        //  FSM_TRANSITION(State1,Event1,State2,Context,Action1), // line A   
        //  ...
        //  FSM_TRANSITION(State1,Event1,State2,Context,Action2), // ERROR: transition from State1 to State2 triggered by Event1 already exists! (conflict with line A)
        //  ...
        //  FSM_TRANSITION(State1,Event1,State3,Context,Action3), // ERROR: transition from State1 triggered by Event1 to other state already exists! (conflict with line A) 
        //  ...
        //  FSM_TRANSITION(State2,Event2,State3,Context,Action4), // line B
        //  FSM_TRANSITION(State2,Event3,State3,Context,Action4), // ERROR: Hmm... you try to go to State3 from State2 by means the different events but with the same action! (conflict with line B) 
        //  FSM_TRANSITION(State2,Event3,State3,Context2,Action4),// CORRECT: contexts are different i.e. you use diffrent actions
        //  ...
        // will be rejected by compiler
        friend void CHECK_DUPLICATION(Loki::Int2Type<BASE_STATE>, Loki::Int2Type<EVENT_VALUE>) {}
        friend void CHECK_DUPLICATION(Loki::Int2Type<BASE_STATE>, Loki::Int2Type<TARGET_STATE>, MemberFunction2Type<ACTION> ) {}
    }; 
};

template
<
    typename TRANSITION_TABLE_T   // is typelist of transitions where a transition is concrete type-instantiation of TransitionT<...>::DefineT<...> template. 
> 
struct ContextEnumeratorT
{
    typedef Loki::Typelist< typename TRANSITION_TABLE_T::Head::ContextT, 
                            typename ContextEnumeratorT<typename TRANSITION_TABLE_T::Tail>::Result
                          > 
            Result;
};

template<> 
struct ContextEnumeratorT<Loki::NullType>
{
    typedef Loki::NullType Result;
};

//////////////////////////////////////////////////////////////////
//  Monitor Default Policy
template
<
    typename STATE_T,
    typename EVENT_T
>
class DefaultMonitor
{
public:
        // event received
    void on_received        (STATE_T currentState, EVENT_T receivedEvent) {}
        // current state was changed
    void on_state_changed   (STATE_T currentState, STATE_T previousState, EVENT_T causalEvent ) {}
        // transition completed i.e. the appropriate action was executed
    void on_completed       (STATE_T currentState, STATE_T previousState, EVENT_T causalEvent ) {}
        // the current state doesn't accept requested event, i.e. there is no transition which can be triggered by given event on given state 
    void unhandled_event    (STATE_T currentState, EVENT_T receivedEvent) {}
};

//////////////////////////////////////////////////////////////////
//  Default Context class 
//  with doNothing() action
struct DefaultContext
{
    void doNothing() {}
};

//////////////////////////////////////////////////////////////////
//  Finite State Machine main template

template
<
    typename TRANSITION_TABLE_T,  // is typelist of transitions where transition is concrete type-instantiation of Transition<...>::DefineT<...> template. 
    typename STATE_T = int,
    typename EVENT_T = int,
    template <class,class> class MONITOR_T = DefaultMonitor
>
class Builder : protected MultipleInheritance               // Builder must be derived from all class-contexts
                        <   typename Loki::TL::NoDuplicates // eliminates class-context duplications
                                                <   typename ContextEnumeratorT<TRANSITION_TABLE_T>::Result 
                                                >::Result
                        >,
             protected MONITOR_T<STATE_T,EVENT_T>

{
    typedef MONITOR_T<STATE_T,EVENT_T>    MONITOR_T;

public:
    Builder() : state_(static_cast<STATE_T>(TRANSITION_TABLE_T::Head::BASE_STATE)) {}

    STATE_T state() const               { return state_; }
    void    postEvent(EVENT_T event)    { 
                                            MONITOR_T::on_received(state_,event);
                                            execTransition<TRANSITION_TABLE_T>(event); 
                                        }

private:
    template <typename TRANSITION_TABLE_ARG_T>
        void execTransition(EVENT_T event)
            {
                typedef typename TRANSITION_TABLE_ARG_T::Head TRANSITION_T;

                if(TRANSITION_T::BASE_STATE == state_ && TRANSITION_T::EVENT == event)
                {
                    state_ = static_cast<STATE_T>(TRANSITION_T::TARGET_STATE);
                    MONITOR_T::on_state_changed(    state_,
                                                    static_cast<STATE_T>(TRANSITION_T::BASE_STATE),
                                                    static_cast<EVENT_T>(TRANSITION_T::EVENT)
                                               );
                    TRANSITION_T::invoke(static_cast<TRANSITION_T::ContextT*>(this));
                    MONITOR_T::on_completed(    state_,
                                                static_cast<STATE_T>(TRANSITION_T::BASE_STATE),
                                                static_cast<EVENT_T>(TRANSITION_T::EVENT)
                                           );
                    return;
                }
                execTransition<typename TRANSITION_TABLE_ARG_T::Tail>(event);
            }

    template <>
        void execTransition<Loki::NullType>(EVENT_T event)
            {
                MONITOR_T::unhandled_event(state_,event);
            }

private:
    STATE_T  state_;
};

}   // end of fsm namespace

#define FSM_TRANSITION(BaseState,Event,TargetState,ContextClass,ContextMethod)   \
    fsm::TransitionT<ContextClass>::DefineT<BaseState,Event,TargetState,ContextClass::ContextMethod>

#define FSM_TRANSITION_NO_ACTION(BaseState,Event,TargetState)   \
    fsm::TransitionT<fsm::DefaultContext>::DefineT<BaseState,Event,TargetState,fsm::DefaultContext::doNothing>

#endif // #ifndef __FINITE_STATE_MACHINE_H__
