#include "logging.h"
#include "thirdPartyComponents.h"
#include <crtdbg.h>
#include <iostream>

#define LogFSM static_cast<LoggingSystem*>(this)

////////////////////////////////////////
// Monitor::unhandled_event implementation
//

void fsm::DefaultMonitor<State,Event>::unhandled_event(State currentState, Event receivedEvent)
{
    _ASSERT(false); // unhandled events are not allowed!!!
    log::trace("!!! Programmer's error: unhandled events are not allowed");
}

////////////////////////////////////////
// Context internal class implementation
//

class Context::GuiProxy
{
public:
    GuiProxy()  { gui::initLibrary(); }    
    ~GuiProxy() { gui::uninitLibrary(); }    
};

class Context::KernelProxy
{
public:
    KernelProxy(): _guard(NULL)  {}    
    void initialize(const char* componentName)
                    {
                        _ASSERT(_guard==NULL);
                        _guard = kernel::Guard::getInstance(componentName);
                        _ASSERT(_guard!=NULL);
                    }
    ~KernelProxy()  { 
                        _ASSERT(_guard!=NULL);
                        if(_guard) 
                            kernel::Guard::releaseInstance(_guard); 
                    }    

    const kernel::Guard* guard() const
                    {
                        _ASSERT(_guard);
                        return _guard;
                    }
private:
    kernel::Guard* _guard;
};

////////////////////////////////////////
// Context class implementation
//

Context::Context()
    :   _gui(*new GuiProxy),
        _kernel(*new KernelProxy),
        _attemptCount(0)
{
}

Context::~Context()
{
    uninit();

    delete &_kernel;
    delete &_gui;
}

void Context::init(const char* componentName)
{
    _ASSERT(componentName != NULL);
    _kernel.initialize(componentName);
}

void Context::uninit()
{
}

void Context::clearScreen()
{
    gui::GreetingDlg().doModal();
    LogFSM->postEvent(E_START);
}

void Context::displayLoginScreen()
{
    gui::LoginScreenDlg dlg;
    if(dlg.doModal()==gui::RC_CANCEL)
        LogFSM->postEvent(E_CANCEL);
    else    // gui::RC_OK
    {
        _pwdUser = dlg._pwd;
        LogFSM->postEvent(E_ENTER);
    }
}

void Context::displayBadPasswordScreen()
{
    LogFSM->postEvent((gui::RC_CANCEL==gui::BadPasswordDlg().doModal())? E_CANCEL : E_START);
}

void Context::displayLockScreen()
{
    gui::LockScreenDlg dlg;
    dlg.doModal();
    _pwdAdmin = dlg._pwd;
    LogFSM->postEvent(E_ENTER);
}

void Context::checkPassword()
{
    if(!_kernel.guard()->checkUserPwd(_pwdUser.c_str()))
    {
        ++_attemptCount;
        LogFSM->postEvent(_attemptCount<MAX_ATTEMPTS? E_PASSWORD_BAD : E_THIRD_PASSWORD_BAD);
    }    
    else
    {
        _attemptCount = 0;
        LogFSM->postEvent(E_PASSWORD_GOOD);
    }
}

void Context::checkAdminPassword()
{
    if(!_kernel.guard()->checkAdminPwd(_pwdAdmin.c_str()))
        LogFSM->postEvent(E_PASSWORD_BAD);    
    else
    {
        _attemptCount = 0;
        LogFSM->postEvent(E_PASSWORD_GOOD);    
    }
}

////////////////////////////////////////
// ProcessLauncher class implementation
//

void ProcessLauncher::init(const char* cmdLine)
{
    _cmdLine = cmdLine;
}

void ProcessLauncher::authenticationSuccess()
{
    if(_cmdLine.empty())
        std::cout << "'default process' started successfully." << std::endl;
    else
        std::cout << "'" << _cmdLine.c_str() << "' started successfully." << std::endl;
}
