#ifndef _STATES_ENENTS_ACTIONS_H_
#define _STATES_ENENTS_ACTIONS_H_

#include "../fsm.h"
#include <string>

enum Event
{
    E_START,
    E_ENTER,
    E_CANCEL,
    E_PASSWORD_GOOD,
    E_PASSWORD_BAD,
    E_THIRD_PASSWORD_BAD
};

enum State
{
    S_START,
    S_GREETING,
    S_LOGIN_IN,
    S_LOGGED_IN,
    S_CHECKING_PASSWORD,
    S_NOTIFYING_PASSWORD_BAD,
    S_CHECKING_ADMIN_PASSWORD,
    S_SCREEN_LOCKED
};

struct Context
{
    void clearScreen();
    void displayLoginScreen();
    void displayBadPasswordScreen();
    void displayLockScreen();

    void checkPassword();
    void checkAdminPassword();

    Context();
    ~Context();
    void init(const char* kernelComponentName);
    void uninit();

private:
    /////////////////////////////////////////////////////////////////
    // implementation of this class will use third party components
    // but including thirdPartyComponents.h is undesirable 
    // potential clients of LoggingSystem do not need to acquire such dependencies.
    // So, we've introduced additional wrappers (indirect relationship layer) GuiProxy and KernelProxy   

    class GuiProxy;
    GuiProxy&       _gui;
    class KernelProxy;
    KernelProxy&    _kernel;    

    std::string     _pwdUser;
    std::string     _pwdAdmin;
    int             _attemptCount;

    enum InternalConstants { MAX_ATTEMPTS = 3 };
};

class ProcessLauncher
{
public:
    void init(const char* cmdLine);
    void authenticationSuccess();

private:
    std::string _cmdLine;
};

typedef TYPELIST_12(
        FSM_TRANSITION(S_START,                   E_START,                S_GREETING,                 Context,           clearScreen),
        FSM_TRANSITION(S_GREETING,                E_START,                S_LOGIN_IN,                 Context,           displayLoginScreen),
        FSM_TRANSITION(S_LOGIN_IN,                E_ENTER,                S_CHECKING_PASSWORD,        Context,           checkPassword),
        FSM_TRANSITION(S_LOGIN_IN,                E_CANCEL,               S_GREETING,                 Context,           clearScreen),
        FSM_TRANSITION(S_CHECKING_PASSWORD,       E_PASSWORD_GOOD,        S_LOGGED_IN,                ProcessLauncher,   authenticationSuccess),
        FSM_TRANSITION(S_CHECKING_PASSWORD,       E_THIRD_PASSWORD_BAD,   S_SCREEN_LOCKED,            Context,           displayLockScreen),
        FSM_TRANSITION(S_CHECKING_PASSWORD,       E_PASSWORD_BAD,         S_NOTIFYING_PASSWORD_BAD,   Context,           displayBadPasswordScreen),
        FSM_TRANSITION(S_NOTIFYING_PASSWORD_BAD,  E_START,                S_LOGIN_IN,                 Context,           displayLoginScreen),
        FSM_TRANSITION(S_NOTIFYING_PASSWORD_BAD,  E_CANCEL,               S_GREETING,                 Context,           clearScreen),
        FSM_TRANSITION(S_SCREEN_LOCKED,           E_ENTER,                S_CHECKING_ADMIN_PASSWORD,  Context,           checkAdminPassword),
        FSM_TRANSITION(S_CHECKING_ADMIN_PASSWORD, E_PASSWORD_BAD,         S_SCREEN_LOCKED,            Context,           displayLockScreen),
        FSM_TRANSITION(S_CHECKING_ADMIN_PASSWORD, E_PASSWORD_GOOD,        S_GREETING,                 Context,           clearScreen)
) 
TransitionTable;


// fsm::DefaultMonitor spesialization to define own unhandled_event implement (see loggin.cpp)
// this specialization will be applied during fsm::Builder<...> template instantiation (see class LoggingSystem below)
template<>
class fsm::DefaultMonitor<State,Event>
{
public:
    void on_received        (State currentState, Event receivedEvent) {}
    void on_state_changed   (State currentState, State previousState, Event overEvent ) {}
    void on_completed       (State currentState, State previousState, Event overEvent ) {}
    void unhandled_event    (State currentState, Event receivedEvent);  // body moved to implementation file to avoid importing of needless dependencies.
};


class LoggingSystem : public fsm::Builder<TransitionTable,State,Event>
{
public:
    LoggingSystem()
        {
            Context::init("kernel_access_rights_validator");
        }
    void setProcessName(const char* name)
        {
            ProcessLauncher::init(name);
        }
};


#endif // !_STATES_ENENTS_ACTIONS_H_